#include "majorminor.h"
#include <string.h>

#define MMCT "MajorMinorCommandTable"

struct subs {
  MajorMinor_Subcommand *vec;
  int len;
};

struct MajorMinor_CommandToken {
  char *name;
  char *message;
  Tcl_HashTable *hash;
  char **namevec;
  struct subs *subcmdvec;
  int subcmdveclen;
  int nameveclen;
  int flags;
};

static void nukeHash _ANSI_ARGS_((ClientData clientData,
				  Tcl_Interp *interp));
static void delcmd   _ANSI_ARGS_((ClientData clientData));
static int MajorminorCommand _ANSI_ARGS_((ClientData clientData,
					  Tcl_Interp *interp, int objc,
					  Tcl_Obj *CONST objv[]));

/* Callback for when the hashtable stored in the interpreter's
 * assocData is nuked.  The contents can be ignored for now. */
static void
nukeHash(ClientData clientData, Tcl_Interp *interp) {
  Tcl_HashTable *hash = (Tcl_HashTable *)clientData;
  Tcl_DeleteHashTable(hash);
  Tcl_Free((char *)hash);
}

/* Callback for when a major/minor command is deleted.  Has to
 * unstitch all the internal references.  Try to guess whether the
 * hashtable got deleted before the command! */
static void
delcmd(ClientData clientData) {
  MajorMinor_CommandToken token = (MajorMinor_CommandToken)clientData;
  Tcl_HashEntry *hent;

  Tcl_Free(token->message);
  Tcl_Free(token->name);

  /* Sanity check - slightly naughty, but useful since I don't know
   * the order of deletion of AssocData and Commands when an
   * interpreter exits. */
  if (token->hash->keyType == TCL_STRING_KEYS) {
    hent = Tcl_FindHashEntry(token->hash, token->name);
    if (hent != NULL) {
      Tcl_DeleteHashEntry(hent);
    }
  }

  if (token->subcmdveclen > 0) {
    int i,j;
    for (i=0 ; i<token->subcmdveclen ; i++) {
      struct subs *s = &token->subcmdvec[i];
      for (j=0 ; j<s->len ; j++) {
	MajorMinor_Subcommand *cmd = &s->vec[j];
	if (cmd->deleteProc != NULL) {
	  cmd->deleteProc(cmd->clientData);
	}
      }
    }
    Tcl_Free((char *)token->subcmdvec);
  }
  if (token->nameveclen > 0) {
    Tcl_Free((char *)token->namevec);
  }
  Tcl_Free((char *)token);
}

/* Handles an actual call to a major/minor command by dispatching off
 * to the correct handling routine. */
static int
MajorminorCommand(ClientData clientData, Tcl_Interp *interp,
		  int objc, Tcl_Obj *CONST objv[]) {
  MajorMinor_CommandToken token = (MajorMinor_CommandToken)clientData;
  int index, i;

  if (objc<2) {
    char *name = Tcl_GetStringFromObj(objv[0], NULL);
    Tcl_AppendResult(interp, "wrong # args: should be \"", name, " ",
		     token->message, " ?arg arg ...?\"", NULL);
    return TCL_ERROR;
  }

  if (token->nameveclen < 1) {
    Tcl_AppendResult(interp, "No ", token->message, " to choose!", NULL);
    return TCL_ERROR;
    /* Or should I panic here? */
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], token->namevec, token->message,
			  token->flags, &index) != TCL_OK) {
    return TCL_ERROR;
  }

  for (i=0 ; i<token->subcmdveclen ; i++) {
    struct subs *subcmds = &token->subcmdvec[i];
    MajorMinor_Subcommand *cmd;

    if (subcmds->len >= index) {
      index -= subcmds->len;
      continue;
    } else {
      cmd = &subcmds->vec[index];
    }

    return cmd->proc(cmd->clientData, interp, objc, objv);
  }

  Tcl_AppendResult(interp, "unexpected fallthrough");
  return TCL_ERROR;
}

/* Look up a command token by name.  Used to get the token to add sets
 * of subcommands to or remove sets of subcommands from.  */
EXTERN MajorMinor_CommandToken
Majorminor_GetCommandToken(Tcl_Interp *interp, char *name) {
  ClientData clientData = Tcl_GetAssocData(interp, MMCT, NULL);
  Tcl_HashEntry *hent;

  if (clientData == NULL) {
    Tcl_AppendResult(interp, "unknown major/minor command \"",
		     name, "\"", NULL);
    return NULL;

  } else {
    Tcl_HashTable *hash = (Tcl_HashTable *)clientData;
    hent = Tcl_FindHashEntry(hash, name);
  }

  if (hent == NULL) {
    Tcl_AppendResult(interp, "unknown major/minor command \"",
		     name, "\"", NULL);
    return NULL;
  }

  return (MajorMinor_CommandToken)Tcl_GetHashValue(hent);
}

/* Create a new command with subcommands.  No subcommands are
 * installed by default, but the token to install them into is
 * returned. */
EXTERN MajorMinor_CommandToken
Majorminor_CreateCommand(Tcl_Interp *interp, char *name,
			 char *msg, int flags) {
  ClientData clientData = Tcl_GetAssocData(interp, MMCT, NULL);
  Tcl_HashTable *hash;
  Tcl_HashEntry *hent;
  int isNew;
  unsigned int len;
  MajorMinor_CommandToken token;

  if (clientData == NULL) {
    hash = (Tcl_HashTable *)Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_SetAssocData(interp, MMCT, &nukeHash, (ClientData)hash);
    Tcl_InitHashTable(hash, TCL_STRING_KEYS);
    hent = Tcl_CreateHashEntry(hash, name, &isNew);

  } else {
    hash = (Tcl_HashTable *)clientData;
    hent = Tcl_CreateHashEntry(hash, name, &isNew);
  }

  if (!isNew) {
    Tcl_AppendResult(interp, "command \"", name, "\" already exists", NULL);
    return NULL;
  }

  token = (MajorMinor_CommandToken)
    Tcl_Alloc(sizeof(struct MajorMinor_CommandToken));
  Tcl_SetHashValue(hent, (ClientData)token);
  token->hash = hash;

  len = strlen(msg)+1;
  token->message = (char *)Tcl_Alloc(len);
  memcpy(token->message, msg, len);

  len = strlen(name)+1;
  token->name = (char *)Tcl_Alloc(len);
  memcpy(token->name, name, len);

  token->flags = flags;

  token->namevec = NULL;
  token->subcmdvec = NULL;
  token->nameveclen = 0;
  token->subcmdveclen = 0;

  Tcl_CreateObjCommand(interp, name, MajorminorCommand, token, delcmd);

  return (MajorMinor_CommandToken)Tcl_GetHashValue(hent);
}

/* Add a set of subcommands to the command given by the token. */
EXTERN void
Majorminor_AddSubcommandSet(MajorMinor_CommandToken command,
			    MajorMinor_Subcommand *subcmdSet) {
  struct subs *sub;
  int olen,i;

  /* Install the block of subcommands.  Note that the block of
   * subcommands itself is assumed to be managed by the calling code.
   * This is reasonable, since you also need to have the code to
   * handle the calls made through the interface, and that is not
   * normally created dynamically... */

  if (command->subcmdveclen == 0) {
    command->subcmdveclen = 1;
    command->subcmdvec = (struct subs *)Tcl_Alloc(sizeof(struct subs)*1);
    sub = &command->subcmdvec[0];
  } else {
    int old = command->subcmdveclen;
    command->subcmdveclen++;
    command->subcmdvec = (struct subs *)
      Tcl_Realloc((char *)command->subcmdvec,
		  sizeof(struct subs)*command->subcmdveclen);
    sub = &command->subcmdvec[old];
  }

  sub->vec = subcmdSet;
  for (sub->len=0 ; sub->vec[sub->len].subcommandName ; sub->len++);

  /* Produce the list of commands. */
  olen = command->nameveclen;
  command->nameveclen += sub->len;
  if (olen > 0) {
    /* This is safe, since we are appending! */
    command->namevec = (char **)
      Tcl_Realloc((char *)command->namevec,
		  sizeof(char *)*command->nameveclen);
  } else {
    command->namevec = (char **)Tcl_Alloc(sizeof(char *)*sub->len);
  }
  for (i=0 ; i<sub->len ; i++,olen++) {
    command->namevec[olen] = sub->vec[i].subcommandName;
  }
}

/* Remove a set of subcommands from the command given by the token.
 * The set must have been installed previously (i.e. its address must
 * have been passed to Majorminor_AddSubcommandSet() on this token.) */
EXTERN int
Majorminor_RemoveSubcommandSet(Tcl_Interp *interp,
			       MajorMinor_CommandToken command,
			       MajorMinor_Subcommand *subcmdSet) {
  if (command->subcmdveclen == 0) {
    Tcl_AppendResult(interp, "no subcommands defined", NULL);
    return TCL_ERROR;
  } else if (command->subcmdveclen == 1) {
    if (command->subcmdvec[0].vec != subcmdSet) {
      Tcl_AppendResult(interp, "this set of subcommands not installed",
		       NULL);
      return TCL_ERROR;
    }
    Tcl_Free((char *)command->subcmdvec);
    command->subcmdveclen = 0;
    Tcl_Free((char *)command->namevec);
    command->nameveclen = 0;
    return TCL_OK;
  } else {
    /* More complex case.  Find which of the subcmdSets we are
     * removing. */
    int idx,lenprev,len;
    char **buffer;
    for (idx=lenprev=0 ; command->subcmdvec[idx].vec != subcmdSet ; idx++) {
      lenprev += command->subcmdvec[idx].len;
      if (idx>=command->subcmdveclen) {
	Tcl_AppendResult(interp, "this set of subcommands not installed",
			 NULL);
	return TCL_ERROR;
      }
    }
    len = command->subcmdvec[idx].len;
    /* shuffle down and resize array */
    for (; idx<command->subcmdveclen-1 ; idx++) {
      command->subcmdvec[idx].vec = command->subcmdvec[idx+1].vec;
      command->subcmdvec[idx].len = command->subcmdvec[idx+1].len;
    }
    command->subcmdvec = (struct subs *)
      Tcl_Realloc((char *)command->subcmdvec,
		  sizeof(struct subs)*(--command->subcmdveclen));

    /* Can't use Tcl_Realloc here safely.  _Must_ allocate afresh! */
    buffer = (char **)Tcl_Alloc(sizeof(char *)*(command->nameveclen-len));
    memcpy(buffer, command->namevec, sizeof(char *)*lenprev);
    memcpy(buffer+lenprev, command->namevec+lenprev+len,
	   sizeof(char *)*(command->nameveclen-lenprev-len));
    Tcl_Free((char *)command->namevec);
    command->namevec = buffer;
    command->nameveclen -= len;

    return TCL_OK;
  }
}

/* Not really needed; this stuff gets performed the first time it is
 * needed anyway. */
EXTERN int
Majorminor_Init(Tcl_Interp *interp) {
  /* Might as well create this here... */
  Tcl_HashTable *hash = (Tcl_HashTable *)Tcl_Alloc(sizeof(Tcl_HashTable));
  Tcl_SetAssocData(interp, MMCT, &nukeHash, (ClientData)hash);
  Tcl_InitHashTable(hash, TCL_STRING_KEYS);

  return TCL_OK;
}
